# -*- coding: utf-8 -*-

##### Research Data related to the publication "Paleo-Hydrogeological Modeling to Understand Present-Day Groundwater Salinities in a Low-Lying Coastal Groundwater System (Northwestern Germany)" by Seibert et al. (2023) #####
#####

### Read iMOD-WQ .idf-files & write Python .npy/.npz-files for model outputs (exemplary 'model_output.py' script, to be executed within the folder where the .run file is stored).

#%%
### Import packages
###

import csv
import sys
print(sys.path)
import imod
import numpy as np
import xarray as xr
import pandas as pd
import geopandas as gpd
from datetime import date
import scipy
import scipy.io as sio
import matplotlib.pyplot as plt
from matplotlib import ticker
import copy
import time
import cftime
import os
from shutil import copyfile

#%%
### Folder dirs
###

model_number = 'MODEL_NAME_HERE' # e.g., 'Model_BC_slice_01'
dir_input = '../imod_input'
dir_imod_input = '../imod_input'
dir_output = '../output/npy_arrays'

#%%
### Model discretization
###
 
nrow = 160  # number of rows
ncol = 240  # number of columns
nlay = 120  # number of layers
rows = np.arange(0, nrow)
columns = np.arange(0, ncol)
layers = np.arange(0, nlay)
dz = 2 # vertical cell extent
dx = 500 # horizontal cell extent
dy = dx # horizontal cell extent
Lx = nrow * dx
Ly = ncol * dy
x_start = 360000 # UTM-X start coordinate
x_end = 480000 # UTM-X end coordinate
x_coord = np.arange(x_start + dx*0.5, x_end, dx) # UTM-X coordinates
y_start = 5880000 # UTM-Y start coordinate
y_end = 5960000 # UTM-Y end coordinate
y_coord = np.arange(y_start + dy*0.5, y_end, dy) # UTM-Y coordinates
ztop = 20. # top elevation [masl]
zbot = ztop - (nlay * dz) # bottom elevation [masl]
botm = np.linspace(ztop - 2, zbot, nlay) # bottom elevations
bot = xr.DataArray(
    botm, {"layer": np.arange(1, 1 + nlay)}, ("layer")
)

#%%
### Load base data
###

# load the paleo ground surface elevation grid
GSE_paleo_path = dir_imod_input+'/GSE_paleo.mat' # paleo ground surface elevation model (.mat file)
GSE_paleo = sio.loadmat(GSE_paleo_path)
GSE_paleo = GSE_paleo['holocene_base']

# load the present-day ground surface elevation grid
GSE_present_path = dir_imod_input+'/GSE_present.mat' # present-day ground surface elevation model (.mat file)
GSE_present = sio.loadmat(GSE_present_path)
GSE_present = GSE_present['surface_grid_full']

#%%
### Initiate loops
###

# read stress period information
stp_data = pd.read_excel(dir_input+"/stp_data/*.xlsx") # stress period information, such as stress period length, max./min. mean sea level etc.; compare ./imod_input/stp_data/; executing the iMOD-Python scripts for each time slice model, the corresponding stp_data files were originally created in ./output/stp_data/ subfolders
per_n = len(np.array(stp_data[0]))
stress_per = np.array(stp_data[0])
per_len2 = np.array(stp_data[2])
sim_years = np.array(stp_data[1])
start_year = int(9000 + per_len2[0]/365 - sim_years[0])
sea_levels = np.array(stp_data[3])
msl_m_cen = np.array(stp_data[4])
    
    
#%%
### Identify stress periods
###

# text box style
props = dict(boxstyle='round', facecolor='wheat', alpha=0.8)

counter_x = -1
stress_periods = np.empty([per_n], dtype=int)
list_1=os.listdir(dir_input+"/results/")
file_num1 = len(list_1)
for i in np.arange(0,file_num1):
    dir_nam = list_1[i]
    list_2 = sorted(os.listdir(dir_input+"/results/"+dir_nam+"/"))
    file_num2 = len(list_2)
    if dir_nam == 'conc':
        for j in np.arange(0,file_num2):
            file_name = list_2[j]
            period_1 = int(file_name[5:13])
            if j == 1:
                counter_x = counter_x + 1
                stress_periods[counter_x] = period_1
            if j >1 and stress_periods[counter_x] != period_1:
                counter_x = counter_x + 1
                stress_periods[counter_x] = period_1
    
### Initiate counters and arrays  

step = 0
sim_t = 0 # days
high_heads = np.empty([10000,5], dtype=float)
counter_heads = 0

###
#%% Extract the output for each stress period
###

for q in np.arange(0,per_n):
    per_nam = stress_periods[q]
    step = step + 1
    sim_t = sim_t + per_len2[q]
    simulation_year = sim_years[q]
    end_file = "_a_paleo"
    
    # current geologic model
    if simulation_year <= 8280:
        GSE = GSE_paleo
    else:
        GSE = GSE_present
    
    ###
    ### Read and store the output for each stress period
    ###
     
    ### HEAD Output
    
    # read head output and transform to .npy array
    head = imod.idf.open(dir_input+"/results/head/*"+str(per_nam)+"*.idf")
    head_2 = np.array(head)
    head_3 = np.zeros([120,160,240])
    head_2[head_2>10000] = np.nan
    head_2 = head_2[-1,:,:,:]
    for i in rows:
        head_3[:,i,:] = head_2[:,len(rows)-1-i,:]
    
    # sum up heads for all layers
    head_2d_lateral_max = np.zeros([160,240])
    head_2d_lateral_mean = np.zeros([160,240])
    for j in rows:
        for k in columns:
            head_col = np.zeros([120])
            for i in layers:
                head_col[i] = head_3[i,j,k]
            head_2d_lateral_max[j,k] = np.nanmax(head_col) # kg/m3
            head_2d_lateral_mean[j,k] = np.nanmean(head_col) # kg/m3
    
    # store head .npy array of nth-stress period
    np.savez_compressed(dir_output+"/heads/heads_"+"{:0>4d}".format(simulation_year)+end_file, head_3)
    np.savez_compressed(dir_output+"/heads_max/heads_max_"+"{:0>4d}".format(simulation_year)+end_file, head_2d_lateral_max)
    np.savez_compressed(dir_output+"/heads_mean/heads_mean_"+"{:0>4d}".format(simulation_year)+end_file, head_2d_lateral_mean)
    
    # store final head .npy array
    if np.nanmax(sim_years) <= 9000:
        if q == per_n -1:
            np.savez_compressed(dir_output+"/"+model_number+"_final_heads", head_3)                    
    
    ### CONCENTRATION Output
    
    # read concentration output and transform to .npy array
    conc = imod.idf.open(dir_input+"/results/conc/*"+str(per_nam)+"*.idf")
    conc_2 = np.array(conc)
    conc_3 = np.zeros([120,160,240])
    conc_2[conc_2>10000] = np.nan
    conc_2 = conc_2[-1,:,:,:]
    for i in rows:
        conc_3[:,i,:] = conc_2[:,len(rows)-1-i,:]
       
    # sum up concentrations for all layers
    conc_2d_lateral_max = np.zeros([160,240])
    conc_2d_lateral_mean = np.zeros([160,240])
    conc_2d_lateral_relative = np.zeros([160,240])
    conc_2d_lateral_depth_to_interface_mbgs = np.zeros([160,240])
    conc_2d_lateral_depth_to_interface_masl = np.zeros([160,240])
    conc_2d_lateral_depth_to_interface_masl[conc_2d_lateral_depth_to_interface_masl == 0] = np.nan
    conc_2d_lateral_depth_to_1g_mbgs = np.zeros([160,240])
    conc_2d_lateral_depth_to_1g_masl = np.zeros([160,240])
    conc_2d_lateral_depth_to_1g_masl[conc_2d_lateral_depth_to_1g_masl == 0] = np.nan
    conc_2d_lateral_depth_to_5g_mbgs = np.zeros([160,240])
    conc_2d_lateral_depth_to_5g_masl = np.zeros([160,240])
    conc_2d_lateral_depth_to_5g_masl[conc_2d_lateral_depth_to_5g_masl == 0] = np.nan
    conc_2d_lateral_depth_to_10g_mbgs = np.zeros([160,240])
    conc_2d_lateral_depth_to_10g_masl = np.zeros([160,240])
    conc_2d_lateral_depth_to_10g_masl[conc_2d_lateral_depth_to_10g_masl == 0] = np.nan
    conc_2d_lateral_depth_to_20g_mbgs = np.zeros([160,240])
    conc_2d_lateral_depth_to_20g_masl = np.zeros([160,240])
    conc_2d_lateral_depth_to_20g_masl[conc_2d_lateral_depth_to_20g_masl == 0] = np.nan
    for j in rows:
        for k in columns:
            conc_col = np.zeros([120])
            counter_cell_active = 0
            counter_cell_saline = 0
            counter_depth_to_interface = 0
            counter_depth_to_1g = 0
            counter_depth_to_5g = 0
            counter_depth_to_10g = 0
            counter_depth_to_20g = 0
            flag_saline1 = 0
            flag_saline2 = 0
            flag_saline3 = 0
            flag_saline4 = 0
            flag_saline5 = 0
            for i in layers:
                conc_col[i] = conc_3[i,j,k]
                if ~np.isnan(conc_3[i,j,k]):
                    counter_cell_active += 1
                    if conc_col[i] > 0.450:
                        counter_cell_saline += 1
                # depth of freshwater-saltwater interface (i.e., 250 mg/L Cl)
                if flag_saline1 == 0 and conc_col[i] < 0.450:
                    counter_depth_to_interface += 1
                elif conc_col[i] >= 0.450:
                    flag_saline1 = 1       
                # depth of 1 g/L salt interface
                if flag_saline2 == 0 and conc_col[i] < 1.0:
                    counter_depth_to_1g += 1
                elif conc_col[i] >= 1.0:
                    flag_saline2 = 1
                # depth of 5 g/L salt interface
                if flag_saline3 == 0 and conc_col[i] < 5.0:
                    counter_depth_to_5g += 1
                elif conc_col[i] >= 5.0:
                    flag_saline3 = 1
                # depth of 10 g/L salt interface
                if flag_saline4 == 0 and conc_col[i] < 10.0:
                    counter_depth_to_10g += 1
                elif conc_col[i] >= 10.0:
                    flag_saline4 = 1
                # depth of 20 g/L salt interface
                if flag_saline5 == 0 and conc_col[i] < 20.0:
                    counter_depth_to_20g += 1
                elif conc_col[i] >= 20.0:
                    flag_saline5 = 1
                    
            conc_2d_lateral_max[j,k] = np.nanmax(conc_col) # kg/m3
            conc_2d_lateral_mean[j,k] = np.nanmean(conc_col) # kg/m3
            if counter_cell_active > 0:
                conc_2d_lateral_relative[j,k] = counter_cell_saline/counter_cell_active * 100 # relative amount of salinized cells [%]
            else:
                conc_2d_lateral_relative[j,k] = np.nan # relative amount of salinized cells [%]
            # depth to interface
            if counter_depth_to_interface > 0:
                conc_2d_lateral_depth_to_interface_mbgs[j,k] = counter_depth_to_interface * 2 # depth to freshwater-saltwater interface [mbgs]; note that each cell has a vertical extent of 2 m
            if counter_depth_to_1g > 0:
                conc_2d_lateral_depth_to_1g_mbgs[j,k] = counter_depth_to_1g * 2 # depth to freshwater-saltwater 1g [mbgs]
            if counter_depth_to_5g > 0:
                conc_2d_lateral_depth_to_5g_mbgs[j,k] = counter_depth_to_5g * 2 # depth to freshwater-saltwater 5g [mbgs]
            if counter_depth_to_10g > 0:
                conc_2d_lateral_depth_to_10g_mbgs[j,k] = counter_depth_to_10g * 2 # depth to freshwater-saltwater 10g [mbgs]
            if counter_depth_to_20g > 0:
                conc_2d_lateral_depth_to_20g_mbgs[j,k] = counter_depth_to_20g * 2 # depth to freshwater-saltwater 20g [mbgs]
                
    # store concentration .npy array of nth-stress period
    np.savez_compressed(dir_output+"/concentrations/concentrations_"+"{:0>4d}".format(simulation_year)+end_file, conc_3)
    np.savez_compressed(dir_output+"/concentrations_max/concentrations_max_"+"{:0>4d}".format(simulation_year)+end_file, conc_2d_lateral_max)
    np.savez_compressed(dir_output+"/concentrations_mean/concentrations_mean_"+"{:0>4d}".format(simulation_year)+end_file, conc_2d_lateral_mean)
    np.savez_compressed(dir_output+"/relative_salinization/relative_salinization_"+"{:0>4d}".format(simulation_year)+end_file, conc_2d_lateral_relative)
    np.savez_compressed(dir_output+"/depth_interface_mbgs/depth_interface_mbgs_"+"{:0>4d}".format(simulation_year)+end_file, conc_2d_lateral_depth_to_interface_mbgs)
    np.savez_compressed(dir_output+"/depth_1g_mbgs/depth_1g_mbgs_"+"{:0>4d}".format(simulation_year)+end_file, conc_2d_lateral_depth_to_1g_mbgs)
    np.savez_compressed(dir_output+"/depth_5g_mbgs/depth_5g_mbgs_"+"{:0>4d}".format(simulation_year)+end_file, conc_2d_lateral_depth_to_5g_mbgs)
    np.savez_compressed(dir_output+"/depth_10g_mbgs/depth_10g_mbgs_"+"{:0>4d}".format(simulation_year)+end_file, conc_2d_lateral_depth_to_10g_mbgs)
    np.savez_compressed(dir_output+"/depth_20g_mbgs/depth_20g_mbgs_"+"{:0>4d}".format(simulation_year)+end_file, conc_2d_lateral_depth_to_20g_mbgs)
    
    # store final concentration .npy array
    if np.nanmax(sim_years) <= 9000:
        if q == per_n -1:
            np.savez_compressed(dir_output+"/"+model_number+"_final_concentrations", conc_3)
    
    ### BUDGET Output
    
    # initiate budget counters and arrays
    pos_sum_drn = 0
    neg_sum_drn = 0
    pos_sum_ghb = 0
    neg_sum_ghb = 0
    pos_sum_rch = 0
    neg_sum_rch = 0
    pos_sum_ch = 0
    neg_sum_ch = 0
    pos_sum_wel = 0
    neg_sum_wel = 0
    pos_sum_riv = 0
    neg_sum_riv = 0
    pos_sum_tot = 0
    neg_sum_tot = 0
    pos_sum_dc_dt = 0
    neg_sum_dc_dt = 0
    bud_rch_3 = np.zeros([120,160,240])
    bud_rch_2 = np.zeros([160,240])
    bud_rch_2_in = np.zeros([160,240])
    bud_rch_2_out = np.zeros([160,240])
    bud_drn_3 = np.zeros([120,160,240])
    bud_drn_2 = np.zeros([160,240])
    bud_drn_2_in = np.zeros([160,240])
    bud_drn_2_out = np.zeros([160,240])
    bud_riv_3 = np.zeros([120,160,240])
    bud_riv_2 = np.zeros([160,240])
    bud_riv_2_in = np.zeros([160,240])
    bud_riv_2_out = np.zeros([160,240])
    bud_ghb_3 = np.zeros([120,160,240])
    bud_ghb_2 = np.zeros([160,240])
    bud_ghb_2_in = np.zeros([160,240])
    bud_ghb_2_out = np.zeros([160,240])
    bud_ch_3 = np.zeros([120,160,240])
    bud_ch_2 = np.zeros([160,240])
    bud_ch_2_in = np.zeros([160,240])
    bud_ch_2_out = np.zeros([160,240])
    bud_wel_3 = np.zeros([120,160,240])
    bud_wel_2 = np.zeros([160,240])
    bud_wel_2_in = np.zeros([160,240])
    bud_wel_2_out = np.zeros([160,240])
    bud_dc_dt_3 = np.zeros([120,160,240])
    bud_dc_dt_2 = np.zeros([160,240])
    bud_dc_dt_2_in = np.zeros([160,240])
    bud_dc_dt_2_out = np.zeros([160,240])
    
    # DRN BUDGET
    # check if budget files exist
    if os.path.isfile(dir_input+"/results/bdgdrn/bdgdrn_"+str(per_nam)+"0000_l1.idf"):
        # drain budget
        bud_drn = imod.idf.open(dir_input+"/results/bdgdrn/*"+str(per_nam)+"*.idf")
        bud_drn_3 = np.array(bud_drn)
        bud_drn_3[bud_drn_3==0] = np.nan
        bud_drn_3 = bud_drn_3[-1,:,:,:]
        bud_drn_2 = np.zeros([160,240])
        # sum up bud for all layers
        for j in rows:
            for k in columns:
                bud_col = np.zeros([120])
                for i in layers:
                    bud_col[i] = bud_drn_3[i,j,k]
                    if bud_drn_3[i,j,k] > 0:
                        pos_sum_drn += bud_drn_3[i,j,k]
                    elif bud_drn_3[i,j,k] < 0:
                        neg_sum_drn += bud_drn_3[i,j,k]
                bud_drn_2[j,k] = np.nansum(bud_col) 
                if np.nansum(bud_col) > 0:
                    bud_drn_2_in[j,k] = np.nansum(bud_col)
                elif np.nansum(bud_col) < 0:
                    bud_drn_2_out[j,k] = np.nansum(bud_col)
        bud_drn_2[bud_drn_2==0] = np.nan
        
        # store budget .npy arrays of nth-stress period
        np.savez_compressed(dir_output+"/budget_drn/budget_drn_3D_"+"{:0>4d}".format(simulation_year)+end_file, bud_drn_3)
        np.savez_compressed(dir_output+"/budget_drn/budget_drn_2D_"+"{:0>4d}".format(simulation_year)+end_file, bud_drn_2)
        np.savez_compressed(dir_output+"/budget_drn/budget_drn_2D_in_"+"{:0>4d}".format(simulation_year)+end_file, bud_drn_2_in)
        np.savez_compressed(dir_output+"/budget_drn/budget_drn_2D_out_"+"{:0>4d}".format(simulation_year)+end_file, bud_drn_2_out)
        np.savez_compressed(dir_output+"/budget_drn/budget_drn_sum_neg_"+"{:0>4d}".format(simulation_year)+end_file, neg_sum_drn)
        np.savez_compressed(dir_output+"/budget_drn/budget_drn_sum_pos_"+"{:0>4d}".format(simulation_year)+end_file, pos_sum_drn)

    # RIV BUDGET
    # check if budget files exist
    if os.path.isfile(dir_input+"/results/bdgriv/bdgriv_"+str(per_nam)+"0000_l1.idf"):
        # river budget
        bud_riv = imod.idf.open(dir_input+"/results/bdgriv/*"+str(per_nam)+"*.idf") 
        bud_riv_3 = np.array(bud_riv)
        bud_riv_3[bud_riv_3==0] = np.nan
        bud_riv_3 = bud_riv_3[-1,:,:,:]
        bud_riv_2 = np.zeros([160,240])
        # sum up bud for all layers
        for j in rows:
            for k in columns:
                bud_col = np.zeros([120])
                for i in layers:
                    bud_col[i] = bud_riv_3[i,j,k]
                    if bud_riv_3[i,j,k] > 0:
                        pos_sum_riv += bud_riv_3[i,j,k]
                    elif bud_riv_3[i,j,k] < 0:
                        neg_sum_riv += bud_riv_3[i,j,k]
                bud_riv_2[j,k] = np.nansum(bud_col)
                if np.nansum(bud_col) > 0:
                    bud_riv_2_in[j,k] = np.nansum(bud_col)
                elif np.nansum(bud_col) < 0:
                    bud_riv_2_out[j,k] = np.nansum(bud_col)
        bud_riv_2[bud_riv_2==0] = np.nan
        
        # store budget .npy arrays of nth-stress period
        np.savez_compressed(dir_output+"/budget_riv/budget_riv_3D_"+"{:0>4d}".format(simulation_year)+end_file, bud_riv_3)
        np.savez_compressed(dir_output+"/budget_riv/budget_riv_2D_"+"{:0>4d}".format(simulation_year)+end_file, bud_riv_2)
        np.savez_compressed(dir_output+"/budget_riv/budget_riv_2D_in_"+"{:0>4d}".format(simulation_year)+end_file, bud_riv_2_in)
        np.savez_compressed(dir_output+"/budget_riv/budget_riv_2D_out_"+"{:0>4d}".format(simulation_year)+end_file, bud_riv_2_out)
        np.savez_compressed(dir_output+"/budget_riv/budget_riv_sum_neg_"+"{:0>4d}".format(simulation_year)+end_file, neg_sum_riv)
        np.savez_compressed(dir_output+"/budget_riv/budget_riv_sum_pos_"+"{:0>4d}".format(simulation_year)+end_file, pos_sum_riv)

    # GHB BUDGET
    # check if budget files exist
    if os.path.isfile(dir_input+"/results/bdgghb/bdgghb_"+str(per_nam)+"0000_l1.idf"):    
        # ghb budget
        bud_ghb = imod.idf.open(dir_input+"/results/bdgghb/*"+str(per_nam)+"*.idf") 
        bud_ghb_3 = np.array(bud_ghb)
        bud_ghb_3[bud_ghb_3==0] = np.nan
        bud_ghb_3 = bud_ghb_3[-1,:,:,:]
        bud_ghb_2 = np.zeros([160,240])
        # sum up bud for all layers
        for j in rows:
            for k in columns:
                bud_col = np.zeros([120])
                for i in layers:
                    bud_col[i] = bud_ghb_3[i,j,k]
                    if bud_ghb_3[i,j,k] > 0:
                        pos_sum_ghb += bud_ghb_3[i,j,k]
                    elif bud_ghb_3[i,j,k] < 0:
                        neg_sum_ghb += bud_ghb_3[i,j,k]
                bud_ghb_2[j,k] = np.nansum(bud_col)
                if np.nansum(bud_col) > 0:
                    bud_ghb_2_in[j,k] = np.nansum(bud_col)
                elif np.nansum(bud_col) < 0:
                    bud_ghb_2_out[j,k] = np.nansum(bud_col)
        bud_ghb_2[bud_ghb_2==0] = np.nan
        
        # store budget .npy arrays of nth-stress period
        np.savez_compressed(dir_output+"/budget_ghb/budget_ghb_3D_"+"{:0>4d}".format(simulation_year)+end_file, bud_ghb_3)
        np.savez_compressed(dir_output+"/budget_ghb/budget_ghb_2D_"+"{:0>4d}".format(simulation_year)+end_file, bud_ghb_2)
        np.savez_compressed(dir_output+"/budget_ghb/budget_ghb_2D_in_"+"{:0>4d}".format(simulation_year)+end_file, bud_ghb_2_in)
        np.savez_compressed(dir_output+"/budget_ghb/budget_ghb_2D_out_"+"{:0>4d}".format(simulation_year)+end_file, bud_ghb_2_out)
        np.savez_compressed(dir_output+"/budget_ghb/budget_ghb_sum_neg_"+"{:0>4d}".format(simulation_year)+end_file, neg_sum_ghb)
        np.savez_compressed(dir_output+"/budget_ghb/budget_ghb_sum_pos_"+"{:0>4d}".format(simulation_year)+end_file, pos_sum_ghb)

    # CH BUDGET
    # check if budget files exist
    if os.path.isfile(dir_input+"/results/bdgbnd/bdgbnd_"+str(per_nam)+"0000_l1.idf"):
        # constant head budget
        bud_ch = imod.idf.open(dir_input+"/results/bdgbnd/*"+str(per_nam)+"*.idf") 
        bud_ch_3 = np.array(bud_ch)
        bud_ch_3[bud_ch_3==0] = np.nan
        bud_ch_3 = bud_ch_3[-1,:,:,:]
        bud_ch_2 = np.zeros([160,240])
        # sum up bud for all layers
        for j in rows:
            for k in columns:
                bud_col = np.zeros([120])
                for i in layers:
                    bud_col[i] = bud_ch_3[i,j,k]
                    if bud_ch_3[i,j,k] > 0:
                        pos_sum_ch += bud_ch_3[i,j,k]
                    elif bud_ch_3[i,j,k] < 0:
                        neg_sum_ch += bud_ch_3[i,j,k]
                bud_ch_2[j,k] = np.nansum(bud_col)
                if np.nansum(bud_col) > 0:
                    bud_ch_2_in[j,k] = np.nansum(bud_col)
                elif np.nansum(bud_col) < 0:
                    bud_ch_2_out[j,k] = np.nansum(bud_col)
        bud_ch_2[bud_ch_2==0] = np.nan
        
        # store budget .npy arrays of nth-stress period
        np.savez_compressed(dir_output+"/budget_ch/budget_ch_3D_"+"{:0>4d}".format(simulation_year)+end_file, bud_ch_3)
        np.savez_compressed(dir_output+"/budget_ch/budget_ch_2D_"+"{:0>4d}".format(simulation_year)+end_file, bud_ch_2)
        np.savez_compressed(dir_output+"/budget_ch/budget_ch_2D_in_"+"{:0>4d}".format(simulation_year)+end_file, bud_ch_2_in)
        np.savez_compressed(dir_output+"/budget_ch/budget_ch_2D_out_"+"{:0>4d}".format(simulation_year)+end_file, bud_ch_2_out)
        np.savez_compressed(dir_output+"/budget_ch/budget_ch_sum_neg_"+"{:0>4d}".format(simulation_year)+end_file, neg_sum_ch)
        np.savez_compressed(dir_output+"/budget_ch/budget_ch_sum_pos_"+"{:0>4d}".format(simulation_year)+end_file, pos_sum_ch)

    # WEL BUDGET
    # check if budget files exist
    if os.path.isfile(dir_input+"/results/bdgwel/bdgwel_"+str(per_nam)+"0000_l1.idf"):
        # well budget
        bud_wel = imod.idf.open(dir_input+"/results/bdgwel/*"+str(per_nam)+"*.idf") 
        bud_wel_3 = np.array(bud_wel)
        bud_wel_3[bud_wel_3==0] = np.nan
        bud_wel_3 = bud_wel_3[-1,:,:,:]
        bud_wel_2 = np.zeros([160,240])
        # sum up bud for all layers
        for j in rows:
            for k in columns:
                bud_col = np.zeros([120])
                for i in layers:
                    bud_col[i] = bud_wel_3[i,j,k]
                    if bud_wel_3[i,j,k] > 0:
                        pos_sum_wel += bud_wel_3[i,j,k]
                    elif bud_wel_3[i,j,k] < 0:
                        neg_sum_wel += bud_wel_3[i,j,k]
                bud_wel_2[j,k] = np.nansum(bud_col)
                if np.nansum(bud_col) > 0:
                    bud_wel_2_in[j,k] = np.nansum(bud_col)
                elif np.nansum(bud_col) < 0:
                    bud_wel_2_out[j,k] = np.nansum(bud_col)
        bud_wel_2[bud_wel_2==0] = np.nan
        
        # store budget .npy arrays of nth-stress period
        np.savez_compressed(dir_output+"/budget_wel/budget_wel_3D_"+"{:0>4d}".format(simulation_year)+end_file, bud_wel_3)
        np.savez_compressed(dir_output+"/budget_wel/budget_wel_2D_"+"{:0>4d}".format(simulation_year)+end_file, bud_wel_2)
        np.savez_compressed(dir_output+"/budget_wel/budget_wel_2D_in_"+"{:0>4d}".format(simulation_year)+end_file, bud_wel_2_in)
        np.savez_compressed(dir_output+"/budget_wel/budget_wel_2D_out_"+"{:0>4d}".format(simulation_year)+end_file, bud_wel_2_out)
        np.savez_compressed(dir_output+"/budget_wel/budget_wel_sum_neg_"+"{:0>4d}".format(simulation_year)+end_file, neg_sum_wel)
        np.savez_compressed(dir_output+"/budget_wel/budget_wel_sum_pos_"+"{:0>4d}".format(simulation_year)+end_file, pos_sum_wel)

    # RCH BUDGET        
    # check if budget files exist
    if os.path.isfile(dir_input+"/results/bdgrch/bdgrch_"+str(per_nam)+"0000_l1.idf"):
        # recharge budget
        bud_rch = imod.idf.open(dir_input+"/results/bdgrch/*"+str(per_nam)+"*.idf")
        bud_rch_3 = np.array(bud_rch)
        bud_rch_3[bud_rch_3==0] = np.nan
        bud_rch_3 = bud_rch_3[-1,:,:,:]
        bud_rch_2 = np.zeros([160,240])
        # sum up bud for all layers
        for j in rows:
            for k in columns:
                bud_col = np.zeros([120])
                for i in layers:
                    bud_col[i] = bud_rch_3[i,j,k]
                    if bud_rch_3[i,j,k] > 0:
                        pos_sum_rch += bud_rch_3[i,j,k]
                    elif bud_rch_3[i,j,k] < 0:
                        neg_sum_rch += bud_rch_3[i,j,k]
                bud_rch_2[j,k] = np.nansum(bud_col)
                if np.nansum(bud_col) > 0:
                    bud_rch_2_in[j,k] = np.nansum(bud_col)
                elif np.nansum(bud_col) < 0:
                    bud_rch_2_out[j,k] = np.nansum(bud_col)
        bud_rch_2[bud_rch_2==0] = np.nan
        
        # store budget .npy arrays of nth-stress period
        np.savez_compressed(dir_output+"/budget_rch/budget_rch_3D_"+"{:0>4d}".format(simulation_year)+end_file, bud_rch_3)
        np.savez_compressed(dir_output+"/budget_rch/budget_rch_2D_"+"{:0>4d}".format(simulation_year)+end_file, bud_rch_2)
        np.savez_compressed(dir_output+"/budget_rch/budget_rch_2D_in_"+"{:0>4d}".format(simulation_year)+end_file, bud_rch_2_in)
        np.savez_compressed(dir_output+"/budget_rch/budget_rch_2D_out_"+"{:0>4d}".format(simulation_year)+end_file, bud_rch_2_out)
        np.savez_compressed(dir_output+"/budget_rch/budget_rch_sum_neg_"+"{:0>4d}".format(simulation_year)+end_file, neg_sum_rch)
        np.savez_compressed(dir_output+"/budget_rch/budget_rch_sum_pos_"+"{:0>4d}".format(simulation_year)+end_file, pos_sum_rch)
        
    # DC-DT BUDGET
    # check if budget files exist
    if os.path.isfile(dir_input+"/results/dcdt/dcdt_"+str(per_nam)+"0000_l1.idf"):
        # recharge budget
        bud_dc_dt = imod.idf.open(dir_input+"/results/dcdt/*"+str(per_nam)+"*.idf")
        bud_dc_dt_3 = np.array(bud_dc_dt)
        bud_dc_dt_3[bud_dc_dt_3==0] = np.nan
        bud_dc_dt_3 = bud_dc_dt_3[-1,:,:,:]
        bud_dc_dt_2 = np.zeros([160,240])
        # sum up bud for all layers
        for j in rows:
            for k in columns:
                bud_col = np.zeros([120])
                for i in layers:
                    bud_col[i] = bud_dc_dt_3[i,j,k]
                    if bud_dc_dt_3[i,j,k] > 0:
                        pos_sum_dc_dt += bud_dc_dt_3[i,j,k]
                    elif bud_dc_dt_3[i,j,k] < 0:
                        neg_sum_dc_dt += bud_dc_dt_3[i,j,k]
                bud_dc_dt_2[j,k] = np.nansum(bud_col)
                if np.nansum(bud_col) > 0:
                    bud_dc_dt_2_in[j,k] = np.nansum(bud_col)
                elif np.nansum(bud_col) < 0:
                    bud_dc_dt_2_out[j,k] = np.nansum(bud_col)
        bud_dc_dt_2[bud_dc_dt_2==0] = np.nan
                        
        # store budget .npy arrays of nth-stress period
        np.savez_compressed(dir_output+"/budget_dc_dt/budget_dc_dt_3D_"+"{:0>4d}".format(simulation_year)+end_file, bud_dc_dt_3)
        np.savez_compressed(dir_output+"/budget_dc_dt/budget_dc_dt_2D_"+"{:0>4d}".format(simulation_year)+end_file, bud_dc_dt_2)
        np.savez_compressed(dir_output+"/budget_dc_dt/budget_dc_dt_2D_in_"+"{:0>4d}".format(simulation_year)+end_file, bud_dc_dt_2_in)
        np.savez_compressed(dir_output+"/budget_dc_dt/budget_dc_dt_2D_out_"+"{:0>4d}".format(simulation_year)+end_file, bud_dc_dt_2_out)
        np.savez_compressed(dir_output+"/budget_dc_dt/budget_dc_dt_sum_neg_"+"{:0>4d}".format(simulation_year)+end_file, neg_sum_dc_dt)
        np.savez_compressed(dir_output+"/budget_dc_dt/budget_dc_dt_sum_pos_"+"{:0>4d}".format(simulation_year)+end_file, pos_sum_dc_dt)
    
    # TOTAL BUDGET
    bud_total_2D_in = bud_rch_2_in + bud_ghb_2_in + bud_drn_2_in + bud_ch_2_in + bud_riv_2_in + bud_wel_2_in
    bud_total_2D_out = bud_rch_2_out + bud_ghb_2_out + bud_drn_2_out + bud_ch_2_out + bud_riv_2_out + bud_wel_2_out
    bud_total_2D = bud_total_2D_in + bud_total_2D_out
    bud_total_sum_in = np.nansum(pos_sum_ch + pos_sum_drn + pos_sum_ghb + pos_sum_rch + pos_sum_riv + pos_sum_wel)
    bud_total_sum_out = np.nansum(neg_sum_ch + neg_sum_drn + neg_sum_ghb + neg_sum_rch + neg_sum_riv + neg_sum_wel)
    # store budget .npy arrays of nth-stress period
    np.savez_compressed(dir_output+"/budget_tot/budget_tot_2D_"+"{:0>4d}".format(simulation_year)+end_file, bud_total_2D)
    np.savez_compressed(dir_output+"/budget_tot/budget_tot_2D_in_"+"{:0>4d}".format(simulation_year)+end_file, bud_total_2D_in)
    np.savez_compressed(dir_output+"/budget_tot/budget_tot_2D_out_"+"{:0>4d}".format(simulation_year)+end_file, bud_total_2D_out)
    np.savez_compressed(dir_output+"/budget_tot/budget_tot_sum_neg_"+"{:0>4d}".format(simulation_year)+end_file, bud_total_sum_out)
    np.savez_compressed(dir_output+"/budget_tot/budget_tot_sum_pos_"+"{:0>4d}".format(simulation_year)+end_file, bud_total_sum_in)

    ### FURTHER OUTPUT
    
    # sea levels
    if 'sea_levels' in locals():
        np.savez_compressed(dir_output+"/sea_levels/sea_level_"+"{:0>4d}".format(simulation_year)+end_file, sea_levels[q])
        np.savez_compressed(dir_output+"/sea_levels/slr_rate_"+"{:0>4d}".format(simulation_year)+end_file, msl_m_cen[q])
    
    # slice name
    if start_year > 0:
        np.savez_compressed(dir_output+"/slice_names/slice_name_"+"{:0>4d}".format(simulation_year)+end_file, "Slice "+"{:0>2d}".format(int(model_number[29:30])))
    elif start_year == 0:
        np.savez_compressed(dir_output+"/slice_names/slice_name_"+"{:0>4d}".format(simulation_year)+end_file, 'Scenario '+model_number[-4::])
        
    # no. of stress period
    np.savez_compressed(dir_output+"/stress_periods/stress_period_"+"{:0>4d}".format(simulation_year)+end_file, stress_per[q])
        
    print(str(q))
